import asyncio
import operator
import threading
from abc import ABCMeta, abstractmethod, abstractproperty
from datetime import datetime, timedelta
from functools import partial
from warnings import warn


class SuspenderBase(metaclass=ABCMeta):
    """An ABC to manage the callbacks between asyincio and pyepics.


    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator or generator function, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator or generator function, optional
            a generator, list, or similar containing `Msg` objects

    tripped_message : str, optional
        Message to include in the trip notification
    """

    def __init__(self, signal, *, sleep=0, pre_plan=None, post_plan=None, tripped_message=""):
        """ """
        self.RE = None
        self._ev = None
        self._tripped = False
        self._tripped_message = tripped_message
        self._sleep = sleep
        self._lock = threading.Lock()
        self._sig = signal
        self._pre_plan = pre_plan
        self._post_plan = post_plan

    def __repr__(self):
        return "{}({!r}, sleep={}, pre_plan={}, post_plan={}, tripped_message={})".format(  # noqa: UP032
            type(self).__name__,
            self._sig,
            self._sleep,
            self._pre_plan,
            self._post_plan,
            self._tripped_message,
        )

    def install(self, RE, *, event_type=None):
        """Install callback on signal

        This (re)installs the required callbacks at the pyepics level

        Parameters
        ----------

        RE : RunEngine
            The run engine instance this should work on

        event_type : str, optional
            The event type (subscription type) to watch
        """
        with self._lock:
            self.RE = RE
        self._sig.subscribe(self, event_type=event_type, run=True)

    def remove(self):
        """Disable the suspender

        Removes the callback at the pyepics level
        """
        self._sig.clear_sub(self)
        with self._lock:
            if self.RE is not None:
                self.__set_event(self.RE._loop)
            self.RE = None
            self._tripped = False

    @abstractmethod
    def _should_suspend(self, value):
        """
        Determine if the current value of the signal is such
        that we need to tell the scan to suspend

        Parameters
        ----------
        value : object
            The value to evaluate to determine if we should
            suspend

        Returns
        -------
        suspend : bool
            True means suspend
        """
        raise NotImplementedError()

    @abstractmethod
    def _should_resume(self, value):
        """
        Determine if the scan is ready to automatically
        restart.

        Parameters
        ----------
        value : object
            The value to evaluate to determine if we should
            resume

        Returns
        -------
        suspend : bool
            True means resume
        """
        raise NotImplementedError()

    def __call__(self, value, **kwargs):
        """Make the class callable so that we can
        pass it off to the ophyd callback stack.

        This expects the massive blob that comes from ophyd
        """
        with self._lock:
            if self.RE is None:
                return
            loop = self.RE._loop

            if self._should_suspend(value):
                self._tripped = True
                # this does dirty things with internal state
                if self._ev is None and self.RE is not None:
                    self.__make_event()
                    if self._ev is None:
                        raise RuntimeError("Could not create the ")
                    cb = partial(
                        self.RE.request_suspend,
                        self._ev.wait,
                        pre_plan=self._pre_plan,
                        post_plan=self._post_plan,
                        justification=self._get_justification(),
                    )
                    if self.RE.state.is_running:
                        loop.call_soon_threadsafe(cb)
            elif self._should_resume(value):
                self.__set_event(loop)
                self._tripped = False

    def __make_event(self):
        """Make or return the asyncio.Event to use as a bridge."""
        assert self._lock.locked()
        if self._ev is None and self.RE is not None:
            th_ev = threading.Event()

            def really_make_the_event():
                self._ev = asyncio.Event()
                th_ev.set()

            h = self.RE._loop.call_soon_threadsafe(really_make_the_event)
            if not th_ev.wait(0.1):
                h.cancel()
        return self._ev

    def __set_event(self, loop):
        """Notify the event that it can resume"""
        assert self._lock.locked()
        if self._ev:
            ev = self._ev
            sleep = self._sleep

            def local():
                ts = (datetime.now() + timedelta(seconds=sleep)).strftime("%Y-%m-%d %H:%M:%S")
                print(
                    f"Suspender {self!r} reports a return to nominal "
                    f"conditions. Will sleep for {sleep} seconds and then "
                    f"release suspension at {ts}."
                )
                # we can use call_later here because this function
                # is scheduled to be run in the event loop thread
                # by the `call_soon_threadsafe` call just below.
                loop.call_later(sleep, ev.set)

            loop.call_soon_threadsafe(local)
        # clear that we have an event
        self._ev = None

    def get_futures(self):
        """Return a list of futures to wait on.

        This will only work correctly if this suspender is 'installed'
        and watching a signal

        Returns
        -------
        futs : list
            List of futures to wait on

        justification : str
            String explaining why the suspender is tripped
        """
        if not self.tripped:
            return [], ""
        with self._lock:
            return [self.__make_event().wait], self._get_justification()

    @property
    def tripped(self):
        return self._tripped

    def _get_justification(self):
        if not self.tripped:
            return ""

        template = "Suspender of type {} stopped by signal {!r}"
        just = template.format(self.__class__.__name__, self._sig)
        return ": ".join(s for s in (just, self._tripped_message) if s)


class SuspendBoolHigh(SuspenderBase):
    """
    Suspend when a boolean signal goes high; resume when it goes low.

    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def _should_suspend(self, value):
        return bool(value)

    def _should_resume(self, value):
        return not bool(value)

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = f"Signal {self._sig.name} is high"
        return ": ".join(s for s in (just, self._tripped_message) if s)


class SuspendBoolLow(SuspenderBase):
    """
    Suspend when a boolean signal goes low; resume when it goes high.

    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def _should_suspend(self, value):
        return not bool(value)

    def _should_resume(self, value):
        return bool(value)

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = f"Signal {self._sig.name} is low"
        return ": ".join(s for s in (just, self._tripped_message) if s)


class _Threshold(SuspenderBase):
    """
    Private base class for suspenders that watch when a scalar
    signal fall above or below a threshold.  Allow for a possibly different
    threshold to resume.
    """

    def __init__(self, signal, suspend_thresh, *, resume_thresh=None, **kwargs):
        super().__init__(signal, **kwargs)
        self._suspend_thresh = suspend_thresh
        if resume_thresh is None:
            resume_thresh = suspend_thresh
        self._resume_thresh = resume_thresh
        self._validate()

    def _should_suspend(self, value):
        return self._op(value, self._suspend_thresh)

    def _should_resume(self, value):
        return not self._op(value, self._resume_thresh)

    @abstractproperty
    def _op(self):
        pass

    @abstractmethod
    def _validate(self):
        pass


class SuspendFloor(_Threshold):
    """
    Suspend when a scalar falls below a threshold.

    Optionally, the threshold to resume can be set to be greater than the
    threshold to suspend.

    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    suspend_thresh : float
        Suspend if the signal value falls below this value

    resume_thresh : float, optional
        Resume when the signal value rises above this value.  If not
        given set to `suspend_thresh`.  Must be greater than `suspend_thresh`.

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def _validate(self):
        if self._resume_thresh < self._suspend_thresh:
            raise ValueError(
                "Resume threshold must be equal or greater "
                "than suspend threshold, you passed: "
                f"suspend: {self._suspend_thresh}  resume: {self._resume_thresh}"
            )

    @property
    def _op(self):
        return operator.lt

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = (
            f"Signal {self._sig.name} = {self._sig.get()!r} "
            + f"fell below {self._suspend_thresh} "
            + f"and has not yet crossed above {self._resume_thresh}."
        )
        return ": ".join(s for s in (just, self._tripped_message) if s)


class SuspendCeil(_Threshold):
    """
    Suspend when a scalar rises above a threshold.

    Optionally, the threshold to resume can be set to be less than the
    threshold to suspend.

    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    suspend_thresh : float
        Suspend if the signal value falls below this value

    resume_thresh : float, optional
        Resume when the signal value rises above this value.  If not
        given set to `suspend_thresh`.  Must be greater than `suspend_thresh`.

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def _validate(self):
        if self._resume_thresh > self._suspend_thresh:
            raise ValueError(
                "Resume threshold must be equal or less "
                "than suspend threshold, you passed: "
                f"suspend: {self._suspend_thresh}  resume: {self._resume_thresh}"
            )

    @property
    def _op(self):
        return operator.gt

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = (
            f"Signal {self._sig.name} = {self._sig.get()!r} "
            + f"went above {self._suspend_thresh} "
            + f"and has not yet crossed below {self._resume_thresh}."
        )
        return ": ".join(s for s in (just, self._tripped_message) if s)


class _SuspendBandBase(SuspenderBase):
    """
    Private base-class for suspenders based on keeping a scalar inside
    or outside of a band
    """

    def __init__(self, signal, band_bottom, band_top, **kwargs):
        super().__init__(signal, **kwargs)
        if not band_bottom < band_top:
            raise ValueError(
                "The bottom of the band must be strictly "
                "less than the top of the band.\n"
                f"bottom: {band_bottom}\ttop: {band_top}"
            )
        self._bot = band_bottom
        self._top = band_top


class SuspendWhenOutsideBand(_SuspendBandBase):
    """
    Suspend when a scalar signal leaves a given band of values.

    Parameters
    ----------
    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    band_bottom, band_top : float
        The top and bottom of the band.  `band_top` must be
        strictly greater than `band_bottom`.

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def _should_resume(self, value):
        return self._bot < value < self._top

    def _should_suspend(self, value):
        return not (self._bot < value < self._top)

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = "Signal {} = {!r} is outside of the range ({}, {})".format(  # noqa: UP032
            self._sig.name, self._sig.get(), self._bot, self._top
        )
        return ": ".join(s for s in (just, self._tripped_message) if s)


class SuspendInBand(SuspendWhenOutsideBand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn(  # noqa: B028
            "SuspendInBand has been renamed SuspendWhenOutsideBand to make "
            "its meaning more clear. Its behavior has not changed."
        )


class SuspendOutBand(_SuspendBandBase):
    """
    Suspend when a scalar signal enters a given band of values.

    This is mostly here because it is the opposite of `SuspenderInBand`.

    Parameters
    ----------

    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    band_bottom, band_top : float
        The top and bottom of the band.  `band_top` must be
        strictly greater than `band_bottom`.

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to 0

    pre_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects

    post_plan : iterable or iterator, optional
            a generator, list, or similar containing `Msg` objects
    """

    def __init__(self, *args, **kwargs):
        warn("bluesky.suspenders.SuspendOutBand is deprecated.")  # noqa: B028
        super().__init__(*args, **kwargs)

    def _should_resume(self, value):
        return not (self._bot < value < self._top)

    def _should_suspend(self, value):
        return self._bot < value < self._top

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = "Signal {} = {!r} is inside of the range ({}, {})".format(  # noqa: UP032
            self._sig.name, self._sig.get(), self._bot, self._top
        )
        return ": ".join(s for s in (just, self._tripped_message) if s)


class SuspendWhenChanged(SuspenderBase):
    """
    Suspend when the monitored value deviates from the expected.

    Only resume if allowed AND when monitored equals expected.

    Notes
    -----

    This suspender is designed to require bluesky restart if value changes.

    USE CASE:

    :class:`~SuspendWhenChanged()` is useful when ``signal`` is an EPICS enumeration
    (`"mbbo" <https://wiki-ext.aps.anl.gov/epics/index.php/RRM_3-14_Multi-Bit_Binary_Output>`_)
    used with a multi-instrument facility.
    Choices predefined in the mbbo record are the
    names of instruments allowed to control any shared hardware.

    * The ``signal``, set by instrument staff outside of bluesky,
      names which instrument is allowed to control the hardware.
    * Other instruments not matching ``signal`` are expected **not** to
      control the hardware (they could use simulators instead or not operate
      the shared hardware).

    Since a decision of hardware *vs.* simulators is made at the
    time a bluesky session starts and ophyd objects are first created, the
    session needs to be aware immediately if the ``signal`` is changed.
    The default value of ``allow_resume=False`` defends this decision.
    If there is a mechanism engineered to toggle ophyd signals between
    hardware and simulators, one might consider ``allow_resume=True``.


    Parameters
    ----------

    signal : `ophyd.Signal`
        The signal to watch for changes to determine if the
        scan should be suspended

    expected_value : str, float, or int
        RunEngine operations will be suspended when signal deviates
        from this value.  If `None` (default), set to value of
        ``signal`` when object is created.

    allow_resume : bool
        Should RunEngine be allowed to resume once ``signal.value == expected``
        again?  Default value of ``False`` is expected for intended use case.

    sleep : float, optional
        How long to wait in seconds after the resume condition is met
        before marking the event as done.  Defaults to ``0``.

    pre_plan : iterable or callable, optional
       Plan to execute just before suspending. If callable, must
       take no arguments.

    post_plan : iterable or callable, optional
        Plan to execute just before resuming. If callable, must
        take no arguments.

    tripped_message : str, optional
        Message to include in the trip notification


    Examples
    --------

    .. code-block:: python

        # pause if this value changes in our session
        # note: this suspender is designed to require Bluesky restart if value changes
        suspend_instrument_in_use = SuspendWhenChanged(instrument_in_use)
        RE.install_suspender(suspend_instrument_in_use)

    Example EPICS database for APS 2-BM-A and 2-BM-B:

    .. code-block:: text

        record(mbbo, "2bm:instrument_in_use") {
            # instrument team sets this
            # For additional field names, see
            # https://epics.anl.gov/EpicsDocumentation/AppDevManuals/RecordRef/Recordref-25.html#HEADING25-15
            field(DESC, "instrument using beam now")
            field(ZRST, "none")
            field(ONST, "2-BM-A")
            field(TWST, "2-BM-B")
            # THST
            # FRST
            # FVST
            # ...
        }

    NOTE: **Always** make the zero choice (``ZRST``) in the mbbo record to be 'none'.
    This allows the instrument staff to designate that *no* instrument is allowed
    to control the shared hardware.  Start the names of the allowed instruments
    with ``ONST``.

    It is convenient for the multi-instrument facility to make this definition
    in EPICS rather than in a specific bluesky session.  The EPICS value could be
    useful in other contexts of instrument control beyond the realm of bluesky.
    """

    def __init__(
        self,
        signal,
        *,
        expected_value=None,
        allow_resume=False,
        sleep=0,
        pre_plan=None,
        post_plan=None,
        tripped_message="",
        **kwargs,
    ):
        self.expected_value = expected_value or signal.value
        self.allow_resume = allow_resume
        super().__init__(
            signal, sleep=sleep, pre_plan=pre_plan, post_plan=post_plan, tripped_message=tripped_message, **kwargs
        )

    def _should_suspend(self, value):
        return value != self.expected_value

    def _should_resume(self, value):
        return self.allow_resume and value == self.expected_value

    def _get_justification(self):
        if not self.tripped:
            return ""

        just = f'Signal {self._sig.name}, got "{self._sig.get()}", expected "{self.expected_value}"'
        if not self.allow_resume:
            just += '.  "RE.abort()" and then restart session to use new configuration.'
        return ": ".join(s for s in (just, self._tripped_message) if s)



import collections
import dataclasses
import enum
import os
import re
import warnings
from typing import Any, Literal, Optional, Union, cast

import numpy as np
from event_model.documents import EventDescriptor, StreamDatum, StreamResource
from tiled.mimetypes import DEFAULT_ADAPTERS_BY_MIMETYPE
from tiled.structures.array import ArrayStructure, BuiltinDtype, StructDtype

# TODO: Move Consolidator classes into external repo (probably area-detector-handlers) and use the existing
# handler discovery mechanism.
# GitHub Issue: https://github.com/bluesky/bluesky/issues/1740


class StructureFamily(str, enum.Enum):
    array = "array"
    awkward = "awkward"
    container = "container"
    sparse = "sparse"
    table = "table"


class Management(str, enum.Enum):
    external = "external"
    immutable = "immutable"
    locked = "locked"
    writable = "writable"


@dataclasses.dataclass
class Asset:
    data_uri: str
    is_directory: bool
    parameter: Optional[str]
    num: Optional[int] = None
    id: Optional[int] = None


@dataclasses.dataclass
class DataSource:
    structure_family: StructureFamily
    structure: Any
    id: Optional[int] = None
    mimetype: Optional[str] = None
    parameters: dict = dataclasses.field(default_factory=dict)
    assets: list[Asset] = dataclasses.field(default_factory=list)
    management: Management = Management.writable


class ConsolidatorBase:
    """Consolidator of StreamDatums

    A Consolidator consumes documents from RE; it is similar to usual Bluesky Handlers but is designed to work
    with streaming data (received via StreamResource and StreamDatum documents). It composes details (DataSource
    and its Assets) that will go into the Tiled database. Each Consolidator is instantiated per a Stream Resource.

    Tiled Adapters will later use this to read the data, with good random access and bulk access support.

    We put this code into consolidators so that additional, possibly very unusual, formats can be supported by
    users without getting a PR merged into Bluesky or Tiled.

    The CONSOLIDATOR_REGISTRY (see example below) and the Tiled catalog parameter adapters_by_mimetype can be used
    together to support:
        - Ingesting a new mimetype from Bluesky documents and generating DataSource and Asset with appropriate
          parameters (the consolidator's job);
        - Interpreting those DataSource and Asset parameters to do I/O (the adapter's job).

    To implement new Consolidators for other mimetypes, subclass ConsolidatorBase, possibly expand the
    `consume_stream_datum` and `get_data_source` methods, and ensure that the keys of returned `adapter_parameters`
    dictionary matches the expected adapter signature. Declare a set of supported mimetypes to allow validation and
    automated discovery of the subclassed Consolidator.

    Attributes:
    -----------

    supported_mimetypes : set[str]
        a set of mimetypes that can be handled by a derived Consolidator class; raises ValueError if attempted to
        pass Resource documents related to unsupported mimetypes.
    join_method : Literal["stack", "concat"]
        a method to join the data; if "stack", the resulting consolidated dataset is produced by joining all datums
        along a new dimension added on the left, e.g. a stack of tiff images, otherwise -- datums will be appended
        to the end of the existing leftmost dimension, e.g. rows of a table (similarly to concatenation in numpy).
    join_chunks : bool
        if True, the chunking of the resulting dataset will be determined after consolidation, otherwise each part
        is considered to be chunked separately.
    """

    supported_mimetypes: set[str] = {"application/octet-stream"}
    join_method: Literal["stack", "concat"] = "concat"
    join_chunks: bool = True

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        self.mimetype = self.get_supported_mimetype(stream_resource)

        self.data_key = stream_resource["data_key"]
        self.uri = stream_resource["uri"]
        self.assets: list[Asset] = []
        self._sres_parameters = stream_resource["parameters"]

        # Find datum shape and machine dtype
        data_desc = descriptor["data_keys"][self.data_key]
        if None in data_desc["shape"]:
            raise NotImplementedError(f"Consolidator for {self.mimetype} does not support variable-sized data")
        self.datum_shape: tuple[int, ...] = cast(tuple[int, ...], tuple(data_desc["shape"]))
        self.datum_shape = () if self.datum_shape == (1,) and self.join_method == "stack" else self.datum_shape

        # Check that the datum shape is consistent between the StreamResource and the Descriptor
        if multiplier := self._sres_parameters.get("multiplier"):
            self.datum_shape = self.datum_shape or (multiplier,)  # If datum_shape is not set
            if self.datum_shape[0] != multiplier:
                if self.datum_shape[0] == 1:
                    self.datum_shape = (multiplier,) + self.datum_shape[1:]
                else:
                    self.datum_shape = (multiplier,) + self.datum_shape
                    # TODO: Check consistency with chunk_shape

        # Determine the machine data type
        self.data_type: Union[BuiltinDtype, StructDtype]
        dtype_numpy = np.dtype(data_desc.get("dtype_numpy"))  # Falls back to np.dtype("float64") if not set
        if dtype_numpy.kind == "V":
            self.data_type = StructDtype.from_numpy_dtype(dtype_numpy)
        else:
            self.data_type = BuiltinDtype.from_numpy_dtype(dtype_numpy)

        # Set chunk (or partition) shape
        self.chunk_shape = self._sres_parameters.get("chunk_shape", ())
        if any(d <= 0 for d in self.chunk_shape):
            raise ValueError(f"Chunk size in all dimensions must be at least 1: chunk_shape={self.chunk_shape}.")

        # Possibly overwrite the join_method and join_chunks attributes
        self.join_method = self._sres_parameters.get("join_method", self.join_method)
        self.join_chunks = self._sres_parameters.get("join_chunks", self.join_chunks)

        self._num_rows: int = 0  # Number of rows in the Data Source (all rows, includung skips)
        self._seqnums_to_indices_map: dict[int, int] = {}

    @classmethod
    def get_supported_mimetype(cls, sres):
        if sres["mimetype"] not in cls.supported_mimetypes:
            raise ValueError(f"A data source of {sres['mimetype']} type can not be handled by {cls.__name__}.")
        return sres["mimetype"]

    @property
    def shape(self) -> tuple[int, ...]:
        """Native shape of the data stored in assets

        This includes the leading (0th) dimension corresponding to the number of rows (if the join_method is stack)
        including skipped rows, if any. The number of relevant usable data rows may be lower, which is determined
        by the `seq_nums` field of StreamDatum documents."""

        if (self.join_method == "concat") and len(self.datum_shape) > 0:
            return self._num_rows * self.datum_shape[0], *self.datum_shape[1:]

        return self._num_rows, *self.datum_shape

    @property
    def chunks(self) -> tuple[tuple[int, ...], ...]:
        """Explicit (dask-style) specification of chunk sizes

        The produced chunk specification is a tuple of tuples of int that specify the sizes of each chunk in each
        dimension; it is based on the StreamResource parameter `chunk_shape`.

        If `chunk_shape` is an empty tuple -- assume the dataset is stored as a single chunk for all existing and
        new elements. Usually, however, `chunk_shape` is a tuple of int, in which case, we assume fixed-sized
        chunks with at most `chunk_shape[0]` elements (i.e. `_num_rows`); last chunk can be smaller. If chunk_shape
        is a tuple with less than `self.shape` elements -- assume it defines the chunk sizes along the leading
        dimensions.

        If the joining method is "concat", and `join_chunks = False`, the chunking along the leftmost dimensions
        is assumed to be preserved in each appended data point, i.e. consecutive chunks do not join, e.g. for a 1d
        array with chunks (3,3,1), the resulting chunking after 3 repeats is (3,3,1,3,3,1,3,3,1).
        When `join_chunks = True` (default), the chunk size along the leftmost dimension is determined by the
        chunk_shape parameter; this is the case when `join_method == "stack"` well.
        Chunking along the trailing dimensions is always preserved as in the original (single) array.
        """

        def list_summands(A: int, b: int, repeat: int = 1) -> tuple[int, ...]:
            # Generate a list with repeated b summing up to A; append the remainder if necessary
            # e.g. list_summands(13, 3) = [3, 3, 3, 3, 1]
            # if `repeat = n`, n > 1, copy and repeat the entire result n times
            return tuple([b] * (A // b) + ([A % b] if A % b > 0 else [])) * repeat or (0,)

        # If chunk shape is less than or equal to the total shape dimensions, chunk each specified dimension
        # starting from the leading dimension
        if len(self.chunk_shape) <= len(self.shape):
            if (
                self.join_method == "stack"
                or (self.join_method == "concat" and self.join_chunks)
                or len(self.chunk_shape) == 0
            ):
                result = tuple(
                    list_summands(ddim, cdim)
                    for ddim, cdim in zip(self.shape[: len(self.chunk_shape)], self.chunk_shape)
                )
            else:
                result = (
                    list_summands(self.datum_shape[0], self.chunk_shape[0], repeat=self._num_rows),
                    *[
                        list_summands(ddim, cdim)
                        for ddim, cdim in zip(self.shape[1 : len(self.chunk_shape)], self.chunk_shape[1:])
                    ],
                )
            return result + tuple((d,) for d in self.shape[len(self.chunk_shape) :])

        # If chunk shape is longer than the total shape dimensions, raise an error
        else:
            raise ValueError(
                f"The shape of chunks, {self.chunk_shape}, should be less than or equal to the shape of data, "
                f"{self.shape}."
            )

    @property
    def has_skips(self) -> bool:
        """Indicates whether any rows should be skipped when mapping their indices to frame numbers

        This flag is intended to provide a shortcut for more efficient data access when there are no skips, and the
        mapping between indices and seq_nums is straightforward. In other case, the _seqnums_to_indices_map needs
        to be taken into account.
        """
        return self._num_rows > len(self._seqnums_to_indices_map)

    def adapter_parameters(self) -> dict:
        """A dictionary of parameters passed to an Adapter

        These parameters are intended to provide any additional information required to read a data source of a
        specific mimetype, e.g. "path" the path into an HDF5 file or "template" the filename pattern of a TIFF
        sequence.

        This method is to be subclassed as necessary.
        """
        return {}

    def structure(self) -> ArrayStructure:
        return ArrayStructure(
            data_type=self.data_type,
            shape=self.shape,
            chunks=self.chunks,
        )

    def consume_stream_datum(self, doc: StreamDatum):
        """Process a new StreamDatum and update the internal data structure

        This will be called for every new StreamDatum received to account for the new added rows.
        This method _may need_ to be subclassed and expanded depending on a specific mimetype.
        Actions:
          - Parse the fields in a new StreamDatum
          - Increment the number of rows (implemented by the Base class)
          - Keep track of the correspondence between indices and seq_nums (implemented by the Base class)
          - Update the list of assets, including their uris, if necessary
          - Update shape and chunks
        """
        self._num_rows += doc["indices"]["stop"] - doc["indices"]["start"]
        new_seqnums = range(doc["seq_nums"]["start"], doc["seq_nums"]["stop"])
        new_indices = range(doc["indices"]["start"], doc["indices"]["stop"])
        self._seqnums_to_indices_map.update(dict(zip(new_seqnums, new_indices)))

    def get_data_source(self) -> DataSource:
        """Return a DataSource object reflecting the current state of the streamed dataset.

        The returned DataSource is conceptually similar (and can be an instance of) tiled.structures.DataSource. In
        general, it describes associated Assets (filepaths, mimetype) along with their internal data structure
        (array shape, chunks, additional parameters) and should contain all information necessary to read the file.
        """
        return DataSource(
            mimetype=self.mimetype,
            assets=self.assets,
            structure_family=StructureFamily.array,
            structure=self.structure(),
            parameters=self.adapter_parameters(),
            management=Management.external,
        )

    def get_adapter(self, adapters_by_mimetype=None):
        """Return an Adapter suitable for reading the data

        Uses a dictionary mapping of a mimetype to a callable that returns an Adapter instance from_catalog.
        """

        # User-provided adapters take precedence over defaults.
        all_adapters_by_mimetype = collections.ChainMap((adapters_by_mimetype or {}), DEFAULT_ADAPTERS_BY_MIMETYPE)
        adapter_class = all_adapters_by_mimetype[self.mimetype]

        # Mimic the necessary aspects of a tiled node with a namedtuple
        _Node = collections.namedtuple("Node", ["metadata_", "specs"])
        return adapter_class.from_catalog(self.get_data_source(), _Node({}, []), **self.adapter_parameters())

    def update_from_stream_resource(self, stream_resource: StreamResource):
        """Consume an additional related StreamResource document for the same data_key"""

        raise NotImplementedError("This method is not implemented in the base Consolidator class.")

    def validate(self, adapters_by_mimetype=None, fix_errors=False):
        """Validate the Consolidator's state against the expected structure"""

        # User-provided adapters take precedence over defaults.
        all_adapters_by_mimetype = collections.ChainMap((adapters_by_mimetype or {}), DEFAULT_ADAPTERS_BY_MIMETYPE)
        adapter_class = all_adapters_by_mimetype[self.mimetype]

        # Initialize adapter from uris and determine the structure
        uris = [asset.data_uri for asset in self.assets if asset.parameter == "data_uris"]
        structure = adapter_class.from_uris(*uris, **self.adapter_parameters()).structure()

        if self.shape != structure.shape:
            if not fix_errors:
                raise ValueError(f"Shape mismatch: {self.shape} != {structure.shape}")
            else:
                warnings.warn(f"Fixing shape mismatch: {self.shape} -> {structure.shape}", stacklevel=2)
                if self.join_method == "stack":
                    self._num_rows = structure.shape[0]
                    self.datum_shape = structure.shape[1:]
                elif self.join_method == "concat":
                    # Estimate the number of frames_per_event (multiplier)
                    multiplier = 1 if structure.shape[0] % structure.chunks[0][0] else structure.chunks[0][0]
                    self._num_rows = structure.shape[0] // multiplier
                    self.datum_shape = (multiplier,) + structure.shape[1:]

        if self.chunks != structure.chunks:
            if not fix_errors:
                raise ValueError(f"Chunk shape mismatch: {self.chunks} != {structure.chunks}")
            else:
                _chunk_shape = tuple(c[0] for c in structure.chunks)
                warnings.warn(f"Fixing chunk shape mismatch: {self.chunk_shape} -> {_chunk_shape}", stacklevel=2)
                self.chunk_shape = _chunk_shape

        if self.data_type != structure.data_type:
            if not fix_errors:
                raise ValueError(f"dtype mismatch: {self.data_type} != {structure.data_type}")
            else:
                warnings.warn(
                    f"Fixing dtype mismatch: {self.data_type.to_numpy_dtype()} -> {structure.data_type.to_numpy_dtype()}",  # noqa
                    stacklevel=2,
                )
                self.data_type = structure.data_type

        assert self.get_adapter() is not None, "Adapter can not not initialized"


class CSVConsolidator(ConsolidatorBase):
    supported_mimetypes: set[str] = {"text/csv;header=absent"}
    join_method: Literal["stack", "concat"] = "concat"
    join_chunks: bool = False

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        super().__init__(stream_resource, descriptor)
        self.assets.append(Asset(data_uri=self.uri, is_directory=False, parameter="data_uris"))

    def adapter_parameters(self) -> dict:
        return {**self._sres_parameters}


class HDF5Consolidator(ConsolidatorBase):
    supported_mimetypes = {"application/x-hdf5"}

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        super().__init__(stream_resource, descriptor)
        self.assets.append(Asset(data_uri=self.uri, is_directory=False, parameter="data_uris", num=0))
        self.swmr = self._sres_parameters.get("swmr", True)

    def adapter_parameters(self) -> dict:
        """Parameters to be passed to the HDF5 adapter, a dictionary with the keys:

        dataset: list[str] - a path to the dataset within the hdf5 file represented as list split at `/`
        swmr: bool -- True to enable the single writer / multiple readers regime
        """
        params = {"dataset": self._sres_parameters["dataset"], "swmr": self.swmr}
        if slice := self._sres_parameters.get("slice", False):
            params["slice"] = slice
        if squeeze := self._sres_parameters.get("squeeze", False):
            params["squeeze"] = squeeze

        return params

    def update_from_stream_resource(self, stream_resource: StreamResource):
        """Add an Asset for a new StreamResource document"""
        if stream_resource["parameters"]["dataset"] != self._sres_parameters["dataset"]:
            raise ValueError("All StreamResource documents must have the same dataset path.")
        if stream_resource["parameters"].get("chunk_shape", ()) != self._sres_parameters.get("chunk_shape", ()):
            raise ValueError("All StreamResource documents must have the same chunk shape.")

        asset = Asset(
            data_uri=stream_resource["uri"], is_directory=False, parameter="data_uris", num=len(self.assets)
        )
        self.assets.append(asset)


class MultipartRelatedConsolidator(ConsolidatorBase):
    def __init__(
        self, permitted_extensions: set[str], stream_resource: StreamResource, descriptor: EventDescriptor
    ):
        super().__init__(stream_resource, descriptor)
        self.permitted_extensions: set[str] = permitted_extensions
        self.data_uris: list[str] = []
        self.chunk_shape = self.chunk_shape or (1,)  # I.e. number of frames per file (tiff, jpeg, etc.)
        if self.join_method == "concat":
            assert self.datum_shape[0] % self.chunk_shape[0] == 0, (
                f"Number of frames per file ({self.chunk_shape[0]}) must divide the total number of frames per "
                f"datum ({self.datum_shape[0]}): variable-sized files are not allowed."
            )

        def int_replacer(match):
            """Normalize filename template

            Replace an integer format specifier with a new-style format specifier, i.e. convert the template string
            from "old" to "new" Python style, e.g. "%s%s_%06d.tif" to "filename_{:06d}.tif"

            """
            flags, width, precision, type_char = match.groups()

            # Handle the flags
            flag_str = ""
            if "-" in flags:
                flag_str = "<"  # Left-align
            if "+" in flags:
                flag_str += "+"  # Show positive sign
            elif " " in flags:
                flag_str += " "  # Space before positive numbers
            if "0" in flags:
                flag_str += "0"  # Zero padding

            # Build width and precision if they exist
            width_str = width if width else ""
            precision_str = f".{precision}" if precision else ""

            # Handle cases like "%6.6d", which should be converted to "{:06d}"
            if precision and width:
                flag_str = "0"
                precision_str = ""
                width_str = str(max(precision, width))

            # Construct the new-style format specifier
            return f"{{:{flag_str}{width_str}{precision_str}{type_char}}}"

        self.template = (
            self._sres_parameters["template"]
            .replace("%s", "{:s}", 1)
            .replace("%s", "")
            .replace("{:s}", self._sres_parameters.get("filename", ""), 1)
        )
        self.template = re.sub(r"%([-+#0 ]*)(\d+)?(?:\.(\d+))?([d])", int_replacer, self.template)

    def get_datum_uri(self, indx: int):
        """Return a full uri for a datum (an individual image file) based on its index in the sequence.

        This relies on the `template` parameter passed in the StreamResource, which is a string in the "new"
        Python formatting style that can be evaluated to a file name using the `.format(indx)` method given an
        integer index, e.g. "{:05d}.ext".
        """
        assert os.path.splitext(self.template)[1] in self.permitted_extensions
        return self.uri + self.template.format(indx)

    def consume_stream_datum(self, doc: StreamDatum):
        """Determine the number and names of files from indices of datums and the number of files per datum.

        In the most general case, each file may be a multipage tiff or a stack of images (frames) and a single
        datum may be composed of multiple such files, leading to a total of self.datum_shape[0] frames.
        Since each file necessarily represents a single chunk (tiffs can not be sub-chunked), the number of
        frames per file is equal to the leftmost chunk_shape dimension, self.chunk_shape[0].
        The number of files produced per each datum is then the ratio of these two numbers.

        If `join_method == "stack"`, we assume that each datum becomes its own index in the new leftmost dimension
        of the resulting dataset, and hence corresponds to a single file.
        """

        files_per_datum = self.datum_shape[0] // self.chunk_shape[0] if self.join_method == "concat" else 1
        first_file_indx = doc["indices"]["start"] * files_per_datum
        last_file_indx = doc["indices"]["stop"] * files_per_datum
        for indx in range(first_file_indx, last_file_indx):
            new_datum_uri = self.get_datum_uri(indx)
            new_asset = Asset(
                data_uri=new_datum_uri,
                is_directory=False,
                parameter="data_uris",
                num=len(self.assets) + 1,
            )
            self.assets.append(new_asset)
            self.data_uris.append(new_datum_uri)

        super().consume_stream_datum(doc)


class TIFFConsolidator(MultipartRelatedConsolidator):
    supported_mimetypes = {"multipart/related;type=image/tiff"}

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        super().__init__({".tif", ".tiff"}, stream_resource, descriptor)


class JPEGConsolidator(MultipartRelatedConsolidator):
    supported_mimetypes = {"multipart/related;type=image/jpeg"}

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        super().__init__({".jpeg", ".jpg"}, stream_resource, descriptor)


class NPYConsolidator(MultipartRelatedConsolidator):
    supported_mimetypes = {"multipart/related;type=application/x-npy"}
    join_method: Literal["stack", "concat"] = "stack"

    # NOTE: NPYConsolidator is tailored for tests in databroker with ophyd.sim devices.
    # Use with caution in other settings!

    def __init__(self, stream_resource: StreamResource, descriptor: EventDescriptor):
        # Unlike other image sequence formats (e.g. TIFF) the filename
        # template is hard-coded in the NPY_SEQ handler. We inject it
        # here so that the rest of the processing can be handled
        # generically by ConsolidatorBase.
        stream_resource["parameters"]["template"] = "%s_%d.npy"
        data_key = stream_resource["data_key"]
        datum_shape = descriptor["data_keys"][data_key]["shape"]
        stream_resource["parameters"]["chunk_shape"] = (1, *datum_shape)
        super().__init__({".npy"}, stream_resource, descriptor)


CONSOLIDATOR_REGISTRY = collections.defaultdict(
    lambda: ConsolidatorBase,
    {
        "text/csv;header=absent": CSVConsolidator,
        "application/x-hdf5": HDF5Consolidator,
        "multipart/related;type=image/tiff": TIFFConsolidator,
        "multipart/related;type=image/jpeg": JPEGConsolidator,
        "multipart/related;type=application/x-npy": NPYConsolidator,
    },
)


def consolidator_factory(stream_resource_doc, descriptor_doc):
    consolidator_class = CONSOLIDATOR_REGISTRY[stream_resource_doc["mimetype"]]
    return consolidator_class(stream_resource_doc, descriptor_doc)



import itertools

import numpy as np
import numpy.testing as npt
import pytest

from bluesky.plan_patterns import (
    OuterProductArgsPattern,
    chunk_outer_product_args,
    classify_outer_product_args_pattern,
    outer_product,
)


@pytest.mark.parametrize(
    "args, pattern",
    [
        (("motor1", 1, 5, 10), OuterProductArgsPattern.PATTERN_1),
        (("motor1", 1, 5, 10, "motor2", 2, 10, 20), OuterProductArgsPattern.PATTERN_1),
        (("motor1", 1, 5, 10, "motor2", 2, 10, 20, "motor3", 3, 20, 40), OuterProductArgsPattern.PATTERN_1),
        (("motor1", 1, 5, 10, "motor2", 2, 10, 20, True), OuterProductArgsPattern.PATTERN_2),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, True, "motor3", 3, 20, 40, False),
            OuterProductArgsPattern.PATTERN_2,
        ),
        # Ambiguous cases (24 elements in 'args')
        (
            (
                "motor",
                1,
                7,
                24,
                "motor1",
                1,
                5,
                10,
                "motor2",
                2,
                10,
                20,
                "motor3",
                3,
                5,
                15,
                "jittery_motor1",
                4,
                6,
                50,
                "jittery_motor2",
                6,
                7,
                30,
                "motor",
                6,
                7,
                30,
            ),
            OuterProductArgsPattern.PATTERN_1,
        ),
        (
            (
                "motor",
                1,
                7,
                24,
                "motor1",
                1,
                5,
                10,
                False,
                "motor2",
                2,
                10,
                20,
                True,
                "motor3",
                3,
                5,
                15,
                False,
                "jittery_motor1",
                4,
                6,
                50,
                True,
                "jittery_motor2",
                6,
                7,
                30,
                False,
            ),
            OuterProductArgsPattern.PATTERN_2,
        ),
    ],
)
def test_classify_outer_product_args_pattern(hw, args, pattern):
    """Basic functionality of 'classify_outer_product_args_pattern"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in args])

    p = classify_outer_product_args_pattern(args)
    assert p == pattern, "Pattern is incorrectly identified"


def test_classify_outer_product_args_pattern_fail(hw):
    """Failing cases for `classify_outer_product_args_pattern`"""

    # Wrong number of elements (doesn't fit any pattern)
    args = (hw.motor1, 1, 5, 10, hw.motor2, 2, 10, 20, hw.motor3, 3, 20)
    with pytest.raises(ValueError, match="Wrong number of elements in 'args'"):
        classify_outer_product_args_pattern(args)

    # Missing or extra motors in the 'args' list
    args_list = [
        (hw.motor1, 1, 5, 10, 50, 2, 10, 20, hw.motor3, 3, 20, 40),
        (hw.motor1, 1, 5, 10, hw.motor2, hw.motor, 10, 20, hw.motor3, 3, 20, 40),
        (hw.motor1, 1, 5, 10, 50, 2, 10, 20, True, hw.motor3, 3, 20, 40, False),
        (hw.motor1, 1, 5, 10, hw.motor2, hw.motor, 10, 20, True, hw.motor3, 3, 20, 40, False),
    ]
    for args in args_list:
        with pytest.raises(ValueError, match="Incorrect order of elements in the argument list 'args'"):
            classify_outer_product_args_pattern(args)


@pytest.mark.parametrize(
    "args, chunked_args",
    [
        (("motor1", 1, 5, 10), [("motor1", 1, 5, 10, False)]),
        (("motor1", 1, 5, 10, "motor2", 2, 10, 20), [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, False)]),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, "motor3", 3, 5, 15),
            [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, False), ("motor3", 3, 5, 15, False)],
        ),
    ],
)
def test_chunk_outer_product_args_1(hw, args, chunked_args):
    """Check if the new pattern (Pattern 1) works"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in args])
    for n, a in enumerate(chunked_args):
        chunked_args[n] = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in a])

    assert list(chunk_outer_product_args(args)) == chunked_args, "Argument list was split into chunks incorrectly"


@pytest.mark.parametrize(
    "args, chunked_args",
    [
        (("motor1", 1, 5, 10), [("motor1", 1, 5, 10, False)]),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, True),
            [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, True)],
        ),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, True, "motor3", 3, 5, 15, False),
            [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, True), ("motor3", 3, 5, 15, False)],
        ),
    ],
)
def test_chunk_outer_product_args_2(hw, args, chunked_args):
    """Check if the pattern (Pattern 2) works"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in args])
    for n, a in enumerate(chunked_args):
        chunked_args[n] = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in a])

    """Check if deprecated pattern (Pattern 2) is properly supported"""
    assert list(chunk_outer_product_args(args)) == chunked_args, "Argument list was split into chunks incorrectly"


@pytest.mark.parametrize(
    "args, chunked_args",
    [
        # Pattern 1
        (
            (
                "motor",
                1,
                7,
                24,
                "motor1",
                1,
                5,
                10,
                "motor2",
                2,
                10,
                20,
                "motor3",
                3,
                5,
                15,
                "jittery_motor1",
                4,
                6,
                50,
                "jittery_motor2",
                6,
                7,
                30,
                "motor",
                6,
                7,
                30,
            ),
            [
                ("motor", 1, 7, 24, False),
                ("motor1", 1, 5, 10, False),
                ("motor2", 2, 10, 20, False),
                ("motor3", 3, 5, 15, False),
                ("jittery_motor1", 4, 6, 50, False),
                ("jittery_motor2", 6, 7, 30, False),
                ("motor", 6, 7, 30, False),
            ],
        ),
        # Pattern 2
        (
            (
                "motor",
                1,
                7,
                24,
                "motor1",
                1,
                5,
                10,
                False,
                "motor2",
                2,
                10,
                20,
                True,
                "motor3",
                3,
                5,
                15,
                False,
                "jittery_motor1",
                4,
                6,
                50,
                True,
                "jittery_motor2",
                6,
                7,
                30,
                False,
            ),
            [
                ("motor", 1, 7, 24, False),
                ("motor1", 1, 5, 10, False),
                ("motor2", 2, 10, 20, True),
                ("motor3", 3, 5, 15, False),
                ("jittery_motor1", 4, 6, 50, True),
                ("jittery_motor2", 6, 7, 30, False),
            ],
        ),
    ],
)
def test_chunk_outer_product_args_3(hw, args, chunked_args):
    """Check the ambiguous case: function is called with 24 arguments,
    the pattern can't be resolved just by counting arguments, so the
    presence of boolean values is checked"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in args])
    for n, a in enumerate(chunked_args):
        chunked_args[n] = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in a])

    assert list(chunk_outer_product_args(args)) == chunked_args, "Argument list was split into chunks incorrectly"


@pytest.mark.parametrize(
    "args, chunked_args, pattern",
    [
        (("motor1", 1, 5, 10), [("motor1", 1, 5, 10, False)], OuterProductArgsPattern.PATTERN_1),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, "motor3", 3, 5, 15),
            [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, False), ("motor3", 3, 5, 15, False)],
            OuterProductArgsPattern.PATTERN_1,
        ),
        (
            ("motor1", 1, 5, 10, "motor2", 2, 10, 20, True, "motor3", 3, 5, 15, False),
            [("motor1", 1, 5, 10, False), ("motor2", 2, 10, 20, True), ("motor3", 3, 5, 15, False)],
            OuterProductArgsPattern.PATTERN_2,
        ),
    ],
)
def test_chunk_outer_product_args_4(hw, args, chunked_args, pattern):
    """Test if `chunk_outer_product_args` works with externally supplied pattern"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in args])
    for n, a in enumerate(chunked_args):
        chunked_args[n] = tuple([getattr(hw, _) if isinstance(_, str) else _ for _ in a])

    assert list(chunk_outer_product_args(args, pattern)) == chunked_args, (
        "Argument list was split into chunks incorrectly"
    )


def test_chunk_outer_product_args_failing(hw):
    """Failing cases for `chunk_outer_product_args` function"""

    # Wrong number of arguments
    args = (hw.motor, 1, 7, 24, True)
    with pytest.raises(ValueError, match="Wrong number of elements in 'args'"):
        list(chunk_outer_product_args(args))

    args = (hw.motor, 1, 7, 24, hw.motor1, 1, 5, 10, False, hw.motor2, 2, 10, 20)
    with pytest.raises(ValueError, match="Wrong number of elements in 'args'"):
        list(chunk_outer_product_args(args))

    args = (
        hw.motor,
        1,
        7,
        24,
        hw.motor1,
        1,
        5,
        10,
        "abc",
        2,
        10,
        20,
    )
    with pytest.raises(ValueError, match="Incorrect order of elements in the argument list"):
        list(chunk_outer_product_args(args))

    args = (hw.motor, 1, 7, 24, hw.motor1, hw.motor3, 5, 10, hw.motor2, 2, 10, 20)
    with pytest.raises(ValueError, match="Incorrect order of elements in the argument list"):
        list(chunk_outer_product_args(args))


def _gen_outer_product(args):
    """
    Generate expected output for the `outer_product` function.
    The output is generated using completely different method not based
    on Cycler, so it is better suited for testing.

    Parameters
    ----------
    args: list
        same as input of the `outer_product` function

    Returns
    -------
    Dictionary:
        {'motor_name_1': list of positions,
         'motor_name_2': list of positions, ...}
    """

    chunk_args = list(chunk_outer_product_args(args))

    # The number of steps is the multiple of the number of steps along all axes
    n_points = 1
    for chunk in chunk_args:
        n_points *= chunk[3]

    positions = {}

    n_passes = 1
    for chunk in chunk_args:
        motor, start, stop, num, snake = chunk

        # The number of measurement during which the motor has to stay at the same position
        n_steps = int(round(n_points / num / n_passes))

        pts = []
        for n in range(n_passes):
            if (n % 2) and snake:  # Odd pass goes backward if the motor is snaked
                v_start, v_stop = stop, start
            else:
                v_start, v_stop = start, stop
            p = list(np.linspace(v_start, v_stop, num))
            p = list(itertools.chain.from_iterable(itertools.repeat(_, n_steps) for _ in p))
            pts += p

        positions[motor.name] = pts

        n_passes *= num

    return positions


@pytest.mark.parametrize(
    "args",
    [
        ("motor", 1, 2, 3, "motor1", 4, 5, 6, True, "motor2", 7, 8, 9, True),
        ("motor", 1, 2, 3, "motor1", 4, 5, 6, True, "motor2", 7, 8, 9, False),
        ("motor", 1, 2, 3, "motor1", 4, 5, 6, False, "motor2", 7, 8, 9, True),
        ("motor", 1, 2, 3, "motor1", 4, 5, 6, True, "motor2", 7, 8, 9, False, "motor3", 7, 8, 9, True),
    ],
)
def test_outer_product(hw, args):
    """Basic test of the functionality"""

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = [getattr(hw, _) if isinstance(_, str) else _ for _ in args]

    full_cycler = outer_product(args=args)
    event_list = list(full_cycler)

    # The list of motors
    chunk_args = list(chunk_outer_product_args(args))
    motors = [_[0] for _ in chunk_args]
    motor_names = [_.name for _ in motors]

    positions = {k: [] for k in motor_names}
    for event in event_list:
        for m, mn in zip(motors, motor_names):
            positions[mn].append(event[m])

    positions_expected = _gen_outer_product(args)

    assert set(positions.keys()) == set(positions_expected.keys()), (
        "Different set of motors in dictionaries of actual and expected positions"
    )

    for name in positions_expected.keys():
        npt.assert_array_almost_equal(
            positions[name],
            positions_expected[name],
            err_msg=f"Expected and actual positions for the motor '{name}' don't match",
        )



import atexit
import logging
import threading
from queue import Empty, Full, Queue
from typing import Callable

logger = logging.getLogger(__name__)


class BufferingWrapper:
    """A wrapper for callbacks that processes documents in a separate thread.

    This class allows a callback to be executed in a background thread, processing
    documents as they are received. This prevent the blocking of RunEngine on any
    slow I/O operations by the callback. It handles graceful shutdown on exit or signal
    termination, ensuring that no new documents are accepted after shutdown has been
    initiated.

    The wrapped callback should be thread-safe and not subscribed to the RE directly.
    If it maintains shared mutable state, it must protect it using internal locking.

    This is mainly a development feature to allow subscribing (potentially many)
    experimental callbacks to a `RunEngine` without the risk of blocking the experiment.
    The use in production is currently not encouraged (at least not without a proper
    testing and risk assessment).

    Parameters
    ----------
        target : callable
            The instance of a callback that will be called with the documents.
            It should accept two parameters: `name` and `doc`.
        queue_size : int, optional
            The maximum size of the internal queue. Default is 1,000,000.

    Usage
    -----
        tw = TiltedWriter(client)
        buff_tw = BufferingWrapper(tw)
        RE.subscribe(buff_tw)
    """

    def __init__(self, target: Callable, queue_size: int = 1_000_000):
        self._wrapped_callback = target
        self._queue: Queue = Queue(maxsize=queue_size)
        self._stop_event = threading.Event()
        self._shutdown_lock = threading.Lock()

        self._thread = threading.Thread(target=self._process_queue, daemon=True)
        self._thread.start()

        atexit.register(self.shutdown)

    def __call__(self, name, doc):
        if self._stop_event.is_set():
            raise RuntimeError("Cannot accept new data after shutdown.")
            # TODO: This can be refactored using the upstream functionality (in Python >= 3.13)
            # https://docs.python.org/3/library/queue.html#queue.Queue.shutdown
        try:
            self._queue.put((name, doc))
        except Full as e:
            logger.exception(
                f"The buffer is full. The {self._wrapped_callback.__class__.__name__} can not keep up with the incoming data. "  # noqa
                f"Consider increasing the queue size or optimizing the callback processing: {e}"
            )
            raise RuntimeError(
                f"The buffer is full. The {self._wrapped_callback.__class__.__name__} can not keep up with the incoming data. "  # noqa
                "Consider increasing the queue size or optimizing the callback processing."
            ) from e
        except Exception as e:
            logger.exception(f"Failed to put document {name} in queue: {e}")
            raise RuntimeError(f"Failed to put document {name} in queue: {e}") from e

    def _process_queue(self):
        while True:
            try:
                if item := self._queue.get(timeout=1):
                    self._wrapped_callback(*item)  # Delegate to wrapped callback
                else:
                    break  # Received sentinel value to stop processing
            except Empty:
                if self._stop_event.is_set():
                    break
            except Exception as e:
                logger.exception(f"Exception in {self._wrapped_callback.__class__.__name__}: {e}")

    def shutdown(self, wait: bool = True):
        if self._stop_event.is_set():
            return
        self._stop_event.set()
        self._queue.put(None)

        atexit.unregister(self.shutdown)

        if wait:
            self._thread.join()

        logger.info(f"{self._wrapped_callback.__class__.__name__} shut down gracefully.")



import collections
import inspect
import random
import re
from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import packaging
import pandas as pd
import pytest

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import bluesky.preprocessors as bpp
from bluesky.plan_patterns import chunk_outer_product_args, outer_product
from bluesky.tests.utils import DocCollector, MsgCollector
from bluesky.utils import RequestStop


def _validate_start(start, expected_values):
    """Basic metadata validtion"""

    plan_md_key = [
        "plan_pattern_module",
        "plan_pattern_args",
        "plan_type",
        "plan_pattern",
        "plan_name",
        "num_points",
        "plan_args",
        "detectors",
    ]

    for k in plan_md_key:
        assert k in start
    for k, v in expected_values.items():
        assert start[k] == v


def test_plan_header(RE, hw):
    args = []

    ##
    args.append(
        (
            bp.grid_scan([hw.det], hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor2, 7, 8, 9, snake_axes=True),
            {
                "motors": ("motor", "motor1", "motor2"),
                "extents": ([1, 2], [4, 5], [7, 8]),
                "shape": (3, 6, 9),
                "snaking": (False, True, True),
                "plan_pattern_module": "bluesky.plan_patterns",
                "plan_pattern": "outer_product",
                "plan_name": "grid_scan",
            },
        )
    )

    ##
    args.append(
        (
            bp.inner_product_scan([hw.det], 9, hw.motor, 1, 2, hw.motor1, 4, 5, hw.motor2, 7, 8),
            {"motors": ("motor", "motor1", "motor2")},
        )
    )

    for plan, target in args:
        c = DocCollector()
        RE(plan, c.insert)
        for s in c.start:
            _validate_start(s, target)


def test_ops_dimension_hints(RE, hw):
    det = hw.det
    motor = hw.motor
    motor1 = hw.motor1
    c = DocCollector()
    RE.subscribe(c.insert)
    RE(bp.grid_scan([det], motor, -1, 1, 7, motor1, 0, 2, 3))

    st = c.start[0]

    assert "dimensions" in st["hints"]

    assert st["hints"]["dimensions"] == [(m.hints["fields"], "primary") for m in (motor, motor1)]


def test_mesh_pseudo(hw, RE):
    p3x3 = hw.pseudo3x3
    sig = hw.sig
    d = DocCollector()

    RE.subscribe(d.insert)
    rs = RE(bp.grid_scan([sig], p3x3.pseudo1, 0, 3, 5, p3x3.pseudo2, 7, 10, 7))

    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    df = pd.DataFrame([_["data"] for _ in d.event[d.descriptor[uid][0]["uid"]]])

    for k in p3x3.describe():
        assert k in df

    for k in sig.describe():
        assert k in df

    assert all(df[sig.name] == 0)
    assert all(df[p3x3.pseudo3.name] == 0)


def test_rmesh_pseudo(hw, RE):
    p3x3 = hw.pseudo3x3
    p3x3.set(1, -2, 100)
    init_pos = p3x3.position
    sig = hw.sig
    d = DocCollector()

    RE.subscribe(d.insert)
    rs = RE(bp.rel_grid_scan([sig], p3x3.pseudo1, 0, 3, 5, p3x3.pseudo2, 7, 10, 7))

    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    df = pd.DataFrame([_["data"] for _ in d.event[d.descriptor[uid][0]["uid"]]])

    for k in p3x3.describe():
        assert k in df

    for k in sig.describe():
        assert k in df

    assert all(df[sig.name] == 0)
    assert all(df[p3x3.pseudo3.name] == 100)
    assert len(df) == 35
    assert min(df[p3x3.pseudo1.name]) == 1
    assert init_pos == p3x3.position


def test_relative_pseudo(hw, RE, db):
    RE.subscribe(db.insert)
    p = hw.pseudo3x3
    p.set(1, 1, 1)
    base_pos = p.position

    # this triggers the merging code path
    rs = RE(bp.relative_inner_product_scan([p], 5, p.pseudo1, -1, 1, p.pseudo2, -2, -1))

    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    tb1 = db[uid].table().drop("time", axis=1)
    assert p.position == base_pos

    # this triggers this does not
    rs = RE(bp.relative_inner_product_scan([p], 5, p.real1, 1, -1, p.real2, 2, 1))
    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]

    tb2 = db[uid].table().drop("time", axis=1)
    assert p.position == base_pos

    # same columns
    assert set(tb1) == set(tb2)
    # same number of points
    assert len(tb1) == len(tb2)

    def get_hint(c):
        h = c.hints["fields"]
        return h[0] if h else c.name

    for c in list(p.pseudo_positioners) + list(p.real_positioners):
        col = get_hint(c)
        print(col)
        assert (tb1[col] == tb2[col]).all()

    assert (tb1[get_hint(p.pseudo1)] == np.linspace(0, 2, 5)).all()


def test_reset_wrapper(hw, RE, monkeypatch):
    monkeypatch.setenv("BLUESKY_PREDECLARE", "1")
    p = hw.pseudo3x3
    m_col = MsgCollector()
    RE.msg_hook = m_col

    RE(bp.relative_inner_product_scan([], 1, p.pseudo1, 0, 1, p.pseudo2, 0, 1))
    expecte_objs = [p, None, None, None, p, None, p, None, None, p, None, None, p, p, None]
    assert len(m_col.msgs) == 15
    assert [m.obj for m in m_col.msgs] == expecte_objs


@pytest.mark.parametrize("pln", [bps.mv, bps.mvr])
def test_pseudo_mv(hw, RE, pln):
    p = hw.pseudo3x3
    m_col = MsgCollector()
    RE.msg_hook = m_col

    RE(pln(p.pseudo1, 1, p.pseudo2, 1))
    expecte_objs = [p, None]
    assert len(m_col.msgs) == 2
    assert [m.obj for m in m_col.msgs] == expecte_objs


def _good_per_step_factory():
    def per_step_old(detectors, motor, step):
        yield from bps.null()

    def per_step_extra(detectors, motor, step, some_kwarg=None):
        yield from bps.null()

    def per_step_exact(detectors, motor, step, take_reading=None):
        yield from bps.null()

    def per_step_kwargs(detectors, motor, step, **kwargs):
        yield from bps.null()

    def per_nd_step(detectors, post_cache, *args, **kwargs):
        yield from bps.null()

    return pytest.mark.parametrize(
        "per_step",
        [per_step_old, per_step_extra, per_step_exact, per_step_kwargs],
        ids=["no kwargs", "extra kwargs", "exact signature", "with kwargs"],
    )


@_good_per_step_factory()
def test_good_per_step_signature(hw, per_step):
    list(bp.scan([hw.det], hw.motor, -1, 1, 5, per_step=per_step))


def _bad_per_step_factory():
    def too_few(detectors, motor):
        "no body"

    def too_many(detectors, motor, step, bob):
        "no body"

    def extra_required_kwarg(detectors, motor, step, *, some_kwarg):
        "no body"

    def wrong_names(a, b, c, take_reading=None):
        "no body"

    def per_step_only_args(*args):
        "no body"

    def per_nd_step_extra(detectors, step, pos_cache, extra_no_dflt):
        "no body"

    def per_nd_step_bad_pos(detectors, step, pos_cache, *, extra_no_dflt):
        "no body"

    def all_wrong(a, b, c=None, *args, d=None, g, **kwargs):
        "no body"

    return pytest.mark.parametrize(
        "per_step",
        [too_few, too_many, extra_required_kwarg, wrong_names, per_step_only_args],
        ids=["too few", "too many", "required kwarg", "bad name", "args only"],
    )


@_bad_per_step_factory()
def test_bad_per_step_signature(hw, per_step):
    sig = inspect.signature(per_step)
    print(f"*** test bad_per_step {sig} ***\n")
    with pytest.raises(
        TypeError,
        match=re.escape(
            "per_step must be a callable with the signature \n "
            "<Signature (detectors, step, pos_cache)> or "
            "<Signature (detectors, motor, step)>. \n"
            f"per_step signature received: {sig}"
        ),
    ):
        list(bp.scan([hw.det], hw.motor, -1, 1, 5, per_step=per_step))


def require_ophyd_1_4_0():
    ophyd = pytest.importorskip("ophyd")
    if packaging.version.Version(ophyd.__version__) < packaging.version.Version("1.4.0"):
        pytest.skip("Needs ophyd 1.4.0 for realistic ophyd.sim Devices.")


@pytest.mark.parametrize("val", [0, None, "aardvark"])
def test_rd_dflt(val):
    ophyd = pytest.importorskip("ophyd")
    require_ophyd_1_4_0()
    sig = ophyd.Signal(value="0", name="sig")

    def tester(obj, dflt):
        ret = yield from bps.rd(obj, default_value=dflt)
        assert ret is dflt

    list(tester(sig, val))


@pytest.mark.parametrize("val", [0, None, "aardvark"])
def test_rd(RE, val):
    ophyd = pytest.importorskip("ophyd")
    require_ophyd_1_4_0()
    sig = ophyd.Signal(value="0", name="sig")

    def tester(obj, val):
        yield from bps.mv(sig, val)
        ret = yield from bps.rd(obj, default_value=object())
        assert ret == val

    RE(tester(sig, val))


def test_rd_fails(hw):
    require_ophyd_1_4_0()
    obj = hw.det

    obj.noise.kind = "hinted"
    hints = obj.hints.get("fields", [])
    msg = re.escape(
        f"Your object {obj} ({obj.name}.{obj.dotted_name}) has {len(hints)} items hinted ({hints}).  We "
    )
    with pytest.raises(ValueError, match=msg):
        list(bps.rd(obj))

    obj.noise.kind = "normal"
    obj.val.kind = "normal"
    msg = re.escape(
        f"Your object {obj} ({obj.name}.{obj.dotted_name}) and has {len(obj.read_attrs)} read attrs.  We "
    )
    with pytest.raises(ValueError, match=msg):
        list(bps.rd(obj))

    obj.read_attrs = []

    msg = re.escape(
        f"Your object {obj} ({obj.name}.{obj.dotted_name}) and has {len(obj.read_attrs)} read attrs.  We "
    )
    with pytest.raises(ValueError, match=msg):
        list(bps.rd(obj))


@pytest.mark.parametrize("kind", ["hinted", "normal"])
def test_rd_device(hw, RE, kind):
    require_ophyd_1_4_0()
    called = False
    hw.det.val.kind = kind

    def tester(obj):
        nonlocal called
        direct_read = yield from bps.read(obj)
        rd_read = yield from bps.rd(obj)
        sig_read = yield from bps.rd(obj.val)

        assert rd_read == direct_read["det"]["value"]
        assert sig_read == rd_read
        called = True

    RE(tester(hw.det))
    assert called


# ********  Tests for `grid_scan` and `rel_grid_scan` plans  ***********


def _retrieve_motor_positions(doc_collector, motors):
    """
    Retrieves the motor positions for the completed run.

    Parameters
    ----------
    `doc_collector`: DocCollector
        DocCollector object that contains data from a single run
    `motors`: list
        the list of motors for which positions should be collected.

    Returns
    -------
    the dictionary:
    {'motor_name_1': list of positions,
     'motor_name_2': list of positions, ...}

    """
    motor_names = [_.name for _ in motors]
    # Get the event list for the first run
    desc = next(iter(doc_collector.event.keys()))  # Descriptor
    event_list = doc_collector.event[desc]

    # Now collect the positions
    positions = {k: [] for k in motor_names}
    for event in event_list:
        for name in positions.keys():
            positions[name].append(event["data"][name])

    return positions


def _grid_scan_position_list(args, snake_axes):
    """
    Generates the lists of positions for each motor during the 'grid_scan'.

    Parameters
    ----------
    args: list
        list of arguments, same as the parameter `args` of the `grid_scan`
    snake_axes: None, True, False or iterable
        same meaning as `snake_axes` parameter of the `grid_scan`

    Returns
    -------
    Tuple of the dictionary:
        {'motor_name_1': list of positions,
         'motor_name_2': list of positions, ...}
    and the tuple that lists `snaking` status of each motor (matches the contents
    of the 'snaking' field of the start document.
    """
    # If 'snake_axis' is specified, it always overwrides any snaking values specified in 'args'
    chunk_args = list(chunk_outer_product_args(args))
    if isinstance(snake_axes, collections.abc.Iterable):
        for n, chunk in enumerate(chunk_args):
            motor, start, stop, num, snake = chunk
            if motor in snake_axes:
                chunk_args[n] = tuple([motor, start, stop, num, True])  # noqa: C409
            else:
                chunk_args[n] = tuple([motor, start, stop, num, False])  # noqa: C409
    elif snake_axes is True:
        chunk_args = [
            (motor, start, stop, num, True) if n > 0 else (motor, start, stop, num, False)
            for n, (motor, start, stop, num, _) in enumerate(chunk_args)
        ]
    elif snake_axes is False:
        chunk_args = [(motor, start, stop, num, False) for (motor, start, stop, num, _) in chunk_args]
    elif snake_axes is None:
        pass
    else:
        raise ValueError(f"The value of 'snake_axes' is not iterable, boolean or None: '{snake_axes}'")

    # Expected contents of the 'snaking' field in the start document
    snaking = tuple([_[4] for _ in chunk_args])

    # Now convert the chunked argument list to regular argument list before calling the cycler
    args_modified = []
    for n, chunk in enumerate(chunk_args):
        if n > 0:
            args_modified.extend(chunk)
        else:
            args_modified.extend(chunk[:-1])

    # Note, that outer_product is used to generate the list of coordinate points
    #   while the plan is executed, but it is tested elsewhere, so it can be trusted
    full_cycler = outer_product(args=args_modified)
    event_list = list(full_cycler)

    # The list of motors
    motors = [_[0] for _ in chunk_args]
    motor_names = [_.name for _ in motors]

    positions = {k: [] for k in motor_names}
    for event in event_list:
        for m, mn in zip(motors, motor_names):
            positions[mn].append(event[m])

    return positions, snaking


@pytest.mark.parametrize(
    "args, snake_axes",
    [
        # Calls using new arguments
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), None),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), False),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), True),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), []),  # Empty list will disable snaking
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), ["motor1"]),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), ["motor2"]),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, "motor2", 7, 8, 9), ["motor1", "motor2"]),
        # Deprecated calls
        (
            ("motor", 1, 2, 3, "motor1", 4, 5, 6, True, "motor2", 7, 8, 9, True),
            None,
        ),  # snake_axes may be only set to None
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, True, "motor2", 7, 8, 9, False), None),
        (("motor", 1, 2, 3, "motor1", 4, 5, 6, False, "motor2", 7, 8, 9, True), None),
    ],
)
@pytest.mark.parametrize("plan, is_relative", [(bp.grid_scan, False), (bp.rel_grid_scan, True)])
def test_grid_scans(RE, hw, args, snake_axes, plan, is_relative):
    """
    Basic test of functionality of `grid_scan` and `rel_grid_scan`:
    Tested:
    - positions of the simulated motors at each step of the scan
    - contents of the 'snaking' field of the start document
    """

    # Convert motor names to actual motors in the argument list using fixture 'hw'
    args = [getattr(hw, _) if isinstance(_, str) else _ for _ in args]
    # Do the same in `snake_axes` if it contains the list of motors
    if isinstance(snake_axes, collections.abc.Iterable):
        snake_axes = [getattr(hw, _) for _ in snake_axes]

    # Place motors at random initial positions. Do it both for relative and
    #   absolute scans. The absolute scans will ignore the inital positions
    #   automatically.
    motors = [_[0] for _ in chunk_outer_product_args(args)]
    motors_pos = [2 * random.random() - 1 for _ in range(len(motors))]
    for _motor, _pos in zip(motors, motors_pos):
        RE(bps.mv(_motor, _pos))

    c = DocCollector()
    RE(plan([hw.det], *args, snake_axes=snake_axes), c.insert)
    positions = _retrieve_motor_positions(c, [hw.motor, hw.motor1, hw.motor2])
    # Retrieve snaking data from the start document
    snaking = c.start[0]["snaking"]

    # Generate the list of positions based on
    positions_expected, snaking_expected = _grid_scan_position_list(args=args, snake_axes=snake_axes)

    assert snaking == snaking_expected, (
        "The contents of the 'snaking' field in the start document does not match the expected values"
    )

    assert set(positions.keys()) == set(positions_expected.keys()), (
        "Different set of motors in dictionaries of actual and expected positions"
    )

    # The dictionary of the initial postiions
    motor_pos_shift = {_motor.name: _pos for (_motor, _pos) in zip(motors, motors_pos)}

    for name in positions_expected.keys():
        # The positions should be shifted only if the plan is relative.
        #   Absolute plans will ignore the initial motor positions
        shift = motor_pos_shift[name] if is_relative else 0
        npt.assert_array_almost_equal(
            positions[name],
            np.array(positions_expected[name]) + shift,
            err_msg=f"Expected and actual positions for the motor '{name}' don't match",
        )


@pytest.mark.parametrize("plan", [bp.grid_scan, bp.rel_grid_scan])
def test_grid_scans_failing(RE, hw, plan):
    """Test the failing cases of 'grid_scan' and 'rel_grid_scan'"""

    # Multiple instance of the same motor in 'args'
    args_list = [
        # New style
        (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor1, 7, 8, 9),
        # Old style
        (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, True, hw.motor1, 7, 8, 9, False),
    ]
    for args in args_list:
        with pytest.raises(ValueError, match="Some motors are listed multiple times in the argument list 'args'"):
            RE(plan([hw.det], *args))

    # 'snake_axes' contains repeated elements
    with pytest.raises(ValueError, match="The list of axes 'snake_axes' contains repeated elements"):
        args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor2, 7, 8, 9)
        snake_axes = [hw.motor1, hw.motor2, hw.motor1]
        RE(plan([hw.det], *args, snake_axes=snake_axes))

    # Snaking is enabled for the slowest motor
    with pytest.raises(ValueError, match="The list of axes 'snake_axes' contains the slowest motor"):
        args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor2, 7, 8, 9)
        snake_axes = [hw.motor1, hw.motor]
        RE(plan([hw.det], *args, snake_axes=snake_axes))

    # Attempt to enable snaking for motors that are not controlled during the scan
    with pytest.raises(
        ValueError,
        match="The list of axes 'snake_axes' contains motors that are not controlled during the scan",
    ):
        args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor2, 7, 8, 9)
        snake_axes = [hw.motor1, hw.motor3]
        RE(plan([hw.det], *args, snake_axes=snake_axes))

    # Mix deprecated and new API ('snake_axes' is used while snaking is set in 'args'
    with pytest.raises(ValueError, match="Mixing of deprecated and new API interface is not allowed"):
        args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, True, hw.motor2, 7, 8, 9, False)
        RE(plan([hw.det], *args, snake_axes=False))

    # The type of 'snake_axes' parameter is not allowed
    for snake_axes in (10, 50.439, "some string"):
        with pytest.raises(ValueError, match="Parameter 'snake_axes' is not iterable, boolean or None"):
            args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, hw.motor2, 7, 8, 9)
            RE(plan([hw.det], *args, snake_axes=snake_axes))

    # Argument detectors is a generator
    generator = (detector for detector in [hw.det1, hw.det2])
    with pytest.raises(
        TypeError, match="The input argument must be either as a list or a tuple of Readable objects."
    ):
        args = (hw.motor, 1, 2, 3, hw.motor1, 4, 5, 6, True, hw.motor2, 7, 8, 9, False)
        RE(plan(generator, *args))


def test_describe_failure(RE):
    ophyd = pytest.importorskip("ophyd")

    class Aardvark(Exception):
        # Unique exception to test behavior of describe.
        ...

    class BadSignalDescribe(ophyd.Signal):
        def describe(self):
            raise Aardvark("Look, an aardvark!")

    class BadSignalRead(ophyd.Signal):
        def read(self):
            raise Aardvark("Look, the other aardvark!")

    bad_signal1 = BadSignalDescribe(value=5, name="Arty")
    bad_signal2 = BadSignalRead(value=5, name="Arty")
    good_signal = ophyd.Signal(value=42, name="baseline")

    class StreamTester:
        def __init__(self):
            self.event_count = 0
            self.stream_names = set()

        def __call__(self, name, doc):
            if name == "event":
                self.event_count += 1
            if name == "descriptor":
                self.stream_names.add(doc["name"])

        def verify(self):
            assert self.event_count == 2
            assert self.stream_names == set(["baseline"])  # noqa: C405

    st = StreamTester()
    with pytest.raises(Aardvark, match="Look, an aardvark!"):
        RE(bpp.baseline_wrapper(bp.count([bad_signal1]), [good_signal]), st)
    st.verify()

    st = StreamTester()
    with pytest.raises(Aardvark, match="Look, the other aardvark!"):
        RE(bpp.baseline_wrapper(bp.count([bad_signal2]), [good_signal]), st)
    st.verify()


def test_errors_through_msg_mutator(hw, monkeypatch):
    monkeypatch.setenv("BLUESKY_PREDECLARE", "1")
    gen = bp.rel_scan([], hw.motor, 5, -5, 10)

    msgs = []

    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(next(gen))
    msgs.append(gen.throw(RequestStop))
    try:
        while True:
            msgs.append(next(gen))
    except RequestStop:
        pass
    else:
        raise False  # noqa: B016

    target = [
        "stage",
        "open_run",
        "declare_stream",
        "checkpoint",
        "set",
        "close_run",
        "unstage",
        "set",
        "wait",
    ]

    assert target == [m.command for m in msgs]


@pytest.mark.parametrize("predeclare", [True, False])
def test_predeclare_env(hw, monkeypatch, predeclare):
    from cycler import cycler

    if predeclare:
        monkeypatch.setenv("BLUESKY_PREDECLARE", "1")

    for p in [
        bp.count([hw.det]),
        bp.scan_nd([hw.det], cycler(hw.motor1, [1, 2, 3]) * cycler(hw.motor2, [4, 5, 6])),
        bp.log_scan([hw.det], hw.motor, 1, 20, 5),
    ]:
        cmds = [m.command for m in p]
        if predeclare:
            assert "declare_stream" in cmds
        else:
            assert "declare_stream" not in cmds


def test_count_failure(RE, hw):
    detector_generator = (det for det in [hw.det1, hw.det2])
    with pytest.raises(TypeError):
        RE(bp.count(detector_generator))

    with pytest.raises(TypeError):
        RE(bp.count(hw.det))


@pytest.mark.filterwarnings("error")
def test_no_warning(hw):
    # this should not warn
    list(bps.rel_set(hw.motor, 2))


def test_input_plan(RE):
    def plan():
        result = yield from bps.input_plan("prompt: ")
        assert result == "answer"

    with patch("bluesky.utils.sys.stdin.readline", return_value="answer\n"):
        RE(plan())



import ophyd

import bluesky
from bluesky.plans import count


def test_blueskyversion(RE):
    assert RE.md["versions"].get("bluesky") == bluesky.__version__


def test_ophydversion(RE):
    assert RE.md["versions"].get("ophyd") == ophyd.__version__


def test_old_md_validator(RE):
    """
    Test an old-style md_validator.

    In older versions of bluesky we ignored the result of md_validator.
    All it could do was raise; it could not provide a normalized valid
    result.

    When an md_validator returns None, we should use the input dict.
    """

    def md_validator(md):
        return None

    start_doc = None

    def test_callback(name, doc):
        nonlocal start_doc
        if name == "start":
            start_doc = doc

    RE.md_validator = md_validator
    RE(count([], md={"test": 1}), test_callback)
    assert start_doc["test"] == 1


def test_md_mormalizer(RE):
    """
    Test the new md_normalizer.

    In current versions of bluesky, md_normalizer is introduced to run
    alongside md_validator. It returns a normalized valid result,
    or raises if it is unable to provide one.

    When an md_normalizer returns a dict, we should use that.
    """

    def md_normalizer(md):
        "Ensure top-level keys are lowercase."
        return {key.lower(): value for key, value in md.items()}

    metadata = {"TEST": 1, "a": {"b": {"c": [1]}}}
    start_doc = None

    def test_callback(name, doc):
        nonlocal start_doc
        if name == "start":
            start_doc = doc

    RE.md_normalizer = md_normalizer
    RE(count([], md=metadata), test_callback)
    assert "TEST" in metadata
    assert "TEST" not in start_doc
    assert "test" in start_doc
    assert start_doc["test"] == 1
    assert start_doc["a"]["b"]["c"] is not metadata["a"]["b"]["c"]



from enum import Enum

import pytest

from bluesky._vendor.super_state_machine import errors, machines


class StatesEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "three"
    FOUR = "four"


class OtherEnum(Enum):
    ONE = "one"


def test_transitions():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = StatesEnum.ONE

        class Meta:
            transitions = {
                StatesEnum.ONE: [StatesEnum.TWO, StatesEnum.THREE],
                StatesEnum.TWO: [StatesEnum.ONE, StatesEnum.THREE],
                StatesEnum.THREE: [StatesEnum.TWO],
            }

    sm = Machine()
    assert sm.is_one is True
    sm.set_two()
    sm.set_three()

    with pytest.raises(errors.TransitionError):
        sm.set_one()
    assert sm.is_three is True

    with pytest.raises(errors.TransitionError):
        sm.set_("one")
    assert sm.is_three is True

    with pytest.raises(errors.TransitionError):
        sm.state = "one"

    assert sm.is_three is True

    sm.set_two()
    sm.set_one()
    assert sm.is_one is True


def test_transitions_checkers():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            transitions = {
                "one": ["two", "three"],
                "two": ["one", "three"],
                "three": ["two", "three"],
            }

    sm = Machine()
    assert sm.can_be_one is False
    assert sm.can_be_two is True
    assert sm.can_be_three is True
    assert sm.can_be_four is False
    assert sm.can_be_("one") is False
    assert sm.can_be_("two") is True
    assert sm.can_be_("three") is True
    assert sm.can_be_("four") is False

    with pytest.raises(errors.TransitionError):
        sm.set_one()

    sm.set_two()
    assert sm.can_be_one is True
    assert sm.can_be_two is False
    assert sm.can_be_three is True
    assert sm.can_be_four is False

    sm.set_three()
    assert sm.can_be_one is False
    assert sm.can_be_two is True
    assert sm.can_be_three is True
    assert sm.can_be_four is False

    sm.set_three()


def test_transitions_checkers_with_complete_graph():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            complete = True
            transitions = {
                "one": ["two", "three"],
                "two": ["one", "three"],
                "three": ["two", "three"],
            }

    sm = Machine()
    sm.set_three()
    assert sm.can_be_one is True
    assert sm.can_be_two is True
    assert sm.can_be_three is True
    assert sm.can_be_four is True
    sm.set_four()
    assert sm.is_four is True


def test_named_transitions_checkers():
    class Machine(machines.StateMachine):
        States = StatesEnum

        class Meta:
            initial_state = "one"
            transitions = {
                "one": ["two", "three"],
                "two": ["one", "three"],
                "three": ["two", "three"],
            }
            named_checkers = [
                ("can_go_to_one", "one"),
                ("can_become_two", StatesEnum.TWO),
            ]

        @property
        def can_one(self):
            return self.can_be_("one")

    sm = Machine()
    sm.set_two()
    assert sm.can_be_one is True
    assert sm.can_one is True
    assert sm.can_go_to_one is True
    assert sm.can_become_two is False


def test_named_transitions_checkers_cant_overwrite_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_checkers = [
                    ("can_one", "one"),
                ]

            @property
            def can_one(self):
                return self.can_be_("one")


def test_named_checkers_cant_overwrite_generated_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_checkers = [
                    ("can_be_one", "one"),
                ]

            @property
            def can_one(self):
                return self.can_be_("one")


def test_named_checkers_dont_accept_wrong_values():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_checkers = [
                    ("can_become_five", "five"),
                ]

            @property
            def can_one(self):
                return self.can_be_("one")


def test_named_checkers_dont_accept_wrong_enums():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_checkers = [
                    ("can_be_other_one", OtherEnum.ONE),
                ]

            @property
            def can_one(self):
                return self.can_be_("one")


def test_transitions_with_wrong_enum():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                transitions = {
                    OtherEnum.ONE: [StatesEnum.TWO, StatesEnum.THREE],
                    StatesEnum.TWO: [StatesEnum.ONE, StatesEnum.THREE],
                    StatesEnum.THREE: [StatesEnum.TWO],
                }


def test_named_transitions():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            transitions = {
                "one": ["two", "three"],
                "two": ["one", "three"],
                "three": ["two", "three"],
            }
            named_transitions = [
                ("run", "two"),
                ("confirm", "three"),
                ("cancel", StatesEnum.FOUR),
            ]

    sm = Machine()
    sm.run()
    assert sm.is_two is True
    sm.confirm()
    assert sm.is_three is True
    sm.cancel()
    assert sm.is_four is True


def test_named_transitions_collisions():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum
            state = "one"

            class Meta:
                transitions = {
                    "o": ["tw", "th"],
                    "tw": ["o", "th"],
                    "th": ["tw"],
                }
                named_transitions = [
                    ("run", "one"),
                    ("confirm", "two"),
                    ("cancel", "three"),
                ]

            def run():
                pass


def test_named_transitions_collisions_with_auto_generated_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum
            state = "one"

            class Meta:
                named_transitions = [
                    ("is_one", "one"),
                    ("confirm", "two"),
                    ("cancel", "three"),
                ]


def test_named_transitions_wrong_value():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum
            state = "one"

            class Meta:
                named_transitions = [
                    ("run", "five"),
                    ("confirm", "two"),
                    ("cancel", "three"),
                ]


def test_named_transitions_wrong_enum():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum
            state = "one"

            class Meta:
                named_transitions = [
                    ("run", OtherEnum.ONE),
                    ("confirm", "two"),
                    ("cancel", "three"),
                ]


def test_named_transitions_are_in_state_graph():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            named_transitions = [
                ("run", "one"),
                ("confirm", "two"),
                ("cancel", "three"),
            ]

    sm = Machine()
    assert sm.can_be_one is True
    assert sm.can_be_two is True
    assert sm.can_be_three is True
    assert sm.can_be_four is False


def test_named_transitions_with_restricted_source():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            named_transitions = [
                ("confirm", "two", "one"),
                ("cancel", "three", ["one", "two"]),
                ("surprise", "four", []),
            ]

    sm = Machine()
    assert sm.can_be_two is True
    assert sm.can_be_three is True
    sm.confirm()
    assert sm.can_be_one is False
    assert sm.can_be_three is True
    assert sm.can_be_four is False
    assert sm.can_be_two is False
    sm.cancel()
    assert sm.is_three is True


def test_restricted_source_proper_value():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_transitions = [
                    ("run", "o", None),
                    ("confirm", "two", "o"),
                    ("cancel", "three", ["o", "tw"]),
                    ("surprise", "four", ["five"]),
                ]


def test_restricted_source_proper_enum():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_transitions = [
                    ("run", "o", None),
                    ("confirm", "two", "o"),
                    ("cancel", "three", ["o", "tw"]),
                    ("surprise", "four", OtherEnum.ONE),
                ]


def test_force_set():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = "one"

        class Meta:
            transitions = {
                "one": ["two", "three"],
            }

    machine = Machine()
    assert machine.can_be_four is False
    with pytest.raises(errors.TransitionError):
        machine.set_four()
    assert machine.state == "one"
    machine.force_set("four")
    assert machine.state == "four"


def test_force_set_name_collision():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            def force_set():
                pass


def test_force_set_name_collision_with_generated_methods():
    with pytest.raises(ValueError):

        class Machine(machines.StateMachine):
            States = StatesEnum

            class Meta:
                named_transitions = [("force_set", "one")]


def test_force_set_accepts_only_proper_values():
    class Machine(machines.StateMachine):
        States = StatesEnum
        state = StatesEnum.ONE

        class Meta:
            complete = False

    machine = Machine()
    assert machine.can_be_four is False
    machine.force_set("four")
    machine.force_set(StatesEnum.FOUR)
    machine.force_set("four")
    with pytest.raises(ValueError):
        machine.force_set("fourtyfour")
    with pytest.raises(ValueError):
        machine.force_set(OtherEnum.ONE)
